/*
 * Copyright (c) 2009, Peter Abeles. All Rights Reserved.
 *
 * This file is part of Efficient Java Matrix Library (EJML).
 *
 * EJML is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 *
 * EJML is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with EJML.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.ejml.alg.dense.decomposition;

import org.ejml.data.DenseMatrix64F;


/**
 * <p>
 * This is an abstract class for singular value decomposition (SVD) algorithms.  It provides
 * common data structures and a generic interface for using the results.  Any child of this
 * class will compute the SVD of A:<br>
 * <div align=center> A = U * W * V <sup>T</sup> </div><br>
 * where A is m by n, and U,W,V are all n by n. U and V are orthogonal matrices.
 * W is a diagonal matrix.
 * </p>
 * 
 * <p>
 * The inverse from SVD is computed as follows:<br>
 *
 * <div align=center>
 * A<sup>-1</sup>=V*[diag(1/w<sub>j</sub>)]*U<sup>T</sup>
 * </div>
 * <br>
 * </p>
 *
 * @see org.ejml.alg.dense.decomposition.svd.SvdNumericalRecipes
 *
 * @author Peter Abeles
 */
public abstract class SingularValueDecomposition implements DecompositionInterface {
    // the shape of the matrix that it can decompose
    // m = number of rows
    // n = number of columns
    protected int m, n;

    // the maximum size of the matrix that it can decompose.
    protected int maxRows=-1,maxCols=-1;

    // used to define how close something needs to be for it to be zero
    protected final static double eps = Math.pow(2.0,-52.0);
    protected double thresh;

    // the refactor matrices
    protected DenseMatrix64F U, V;
    protected double[] u, v, w;

    // data structures used to hold temporary data
    protected double xx[];
    protected double tmp[];

    @Override
    public void setExpectedMaxSize( int numRows , int numCols ) {
        this.maxRows = numRows;
        this.maxCols = numCols;

        U = new DenseMatrix64F(numRows, numCols);
        V = new DenseMatrix64F(numCols, numCols);

        u = U.data;
        w = new double[numCols];
        v = V.data;

        xx = new double[numCols];
        tmp = new double[numCols];
    }

    /**
     * Performs SVD on the specified matrix.  This function must be called first
     * before many of the other functions will function correctly.
     *
     * @param mat The m by n matrix that SVD is to be performed on. Not modified.
     */
    public abstract boolean decompose(DenseMatrix64F mat);

    /**
     * Returns the singular values.  This is the diagonal elements of the W matrix in the decomposition.
     *
     * @return Singular values.
     */
    public double [] getSingularValues() {
        return w;
    }

    /**
     * Returns the 'U' is the refactoring:
     *
     * A = U W V <sup>T</sup> <br>
     *
     * @return An orthogonal n by n matrix.
     */
    public DenseMatrix64F getU() {
        return U;
    }

    /**
     * Returns the 'V' is the refactoring:
     *
     * A = U W V <sup>T</sup> <br>
     *
     * @return An orthogonal n by n matrix.
     */
    public DenseMatrix64F getV() {
        return V;
    }

    /**
     * This changes default threshold that is used to indivate if a value is singular
     * or not.  This should be a small number, around 1e-15.
     *
     * @param thresh The new threshold for singularity.
     */
    public void setSingularThreshold( double thresh ) {
        this.thresh = thresh;
    }

    /**
     * Returns the value that is used to indicate if a value is singular or not.
     *
     * @return Singular threshold.
     */
    public double getSingularThreshold() {
        return thresh;
    }

    /**
     * Compute a reasonable value for the singular value threshold.
     */
    protected void computeDefaultThresh() {
        thresh = 0.5*Math.sqrt(m+n+1.0)*w[0]*eps;
    }

    public int rank() {
        int numRank=0;

        for( int j = 0; j < n; j++ ) {
            if( w[j] > thresh)
                numRank++;
        }

        return numRank;
    }

    public int nullity() {
        int ret = 0;

        for( int j = 0; j < n; j++ ) {
            if( w[j] <= thresh) ret++;
        }
        return ret;
    }

    /**
     *
     *
     *
     * @param rng An m by z matrix where the range space will be writen to. Modified.
     */
    public void range( DenseMatrix64F rng ) {
        int rank = rank();

        if( rng.numRows != m || rng.numCols != rank ) {
            throw new RuntimeException("Invalid shape for rng");
        }

        double dataRng[] = rng.data;

        int nr = 0;
        for( int j =0; j <n; j++ ) {
            if( w[j] > thresh) {
                for( int i =0; i < m; i++ ) {
                    dataRng[i*rank+nr] = u[i*n+j];
                }
                nr++;
            }
        }
    }

    /**
     * The null space of a matrix A is the set of all vectors x for which Ax = 0.
     *
     * The null space has dimension z by m, where z is the nullity of the matrix.
     *
     * @param ns An m by z matrix where the null space will be writen to. Modified.
     */
    public void nullspace( DenseMatrix64F ns ) {
        int nullity = nullity();

        if( ns.numRows != m || ns.numCols != nullity ) {
            throw new RuntimeException("Invalid shape for ns");
        }

        double dataNs[] = ns.data;

        int nr = 0;
        for( int j =0; j <n; j++ ) {
            if( w[j] <= thresh) {
                for( int jj =0; jj < n; jj++ ) {
                    dataNs[jj*nullity+nr] = v[jj*n+j];
                }
                nr++;
            }
        }
    }

    public double[] getW() {
        return w;
    }

    public double[] _getXX() {
        return xx;
    }

    public double[] _getTmp() {
        return tmp;
    }
}
