/*
 * Copyright (c) 2009, Peter Abeles. All Rights Reserved.
 *
 * This file is part of Efficient Java Matrix Library (EJML).
 *
 * EJML is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 *
 * EJML is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with EJML.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.ejml.ops;

import org.ejml.EjmlParameters;
import org.ejml.alg.dense.decomposition.MatrixInvertSpecialized;
import org.ejml.alg.dense.decomposition.lu.LUDecompositionAlt;
import org.ejml.alg.dense.linsol.LinearSolver;
import org.ejml.alg.dense.linsol.lu.LinearSolverLu;
import org.ejml.alg.dense.linsol.qr.LinearSolverQrHouseCol;
import org.ejml.alg.dense.misc.UtilDeterminant;
import org.ejml.alg.dense.mult.MatrixMatrixMult;
import org.ejml.alg.dense.mult.MatrixVectorMult;
import org.ejml.data.DenseMatrix64F;


/**
 * <p>
 * Common matrix operations are contained here.  Which specific underlying algorithm is used
 * is not specified just the out come of the operation.  Nor should calls to these functions
 * reply on the underlying implementation.  Which algorithm is used can depend on the matrix
 * being passed in.
 * </p>
 * <p>
 * For more exotic and specialized generic operations see {@link org.ejml.ops.SpecializedOps}.
 * </p>
 * @see org.ejml.alg.dense.mult.MatrixMatrixMult
 * @see org.ejml.alg.dense.mult.MatrixVectorMult
 * @see org.ejml.ops.SpecializedOps
 * @see org.ejml.ops.MatrixFeatures
 *
 * @author Peter Abeles
 */
@SuppressWarnings({"ForLoopReplaceableByForEach"})
public class CommonOps {
    /**
     * <p>Performs the following operation:<br>
     * <br>
     * c = a * b <br>
     * <br>
     * c<sub>ij</sub> = &sum;<sub>k=1:n</sub> { a<sub>ik</sub> * b<sub>kj</sub>}
     * </p>
     *
     * @param a The left matrix in the multiplcation operation. Not modified.
     * @param b The right matrix in the multiplcation operation. Not modified.
     * @param c Where the results of the operation are stored. Modified.
     */
    public static void mult( DenseMatrix64F a , DenseMatrix64F b , DenseMatrix64F c )
    {
        if( b.numCols >= EjmlParameters.MULT_COLUMN_SWITCH ) {
            MatrixMatrixMult.mult_reorder(a,b,c);
        } else {
            MatrixMatrixMult.mult_small(a,b,c);
        }
    }

    /**
     * <p>Performs the following operation:<br>
     * <br>
     * c = &alpha; * a * b <br>
     * <br>
     * c<sub>ij</sub> = &alpha; &sum;<sub>k=1:n</sub> { * a<sub>ik</sub> * b<sub>kj</sub>}
     * </p>
     *
     * @param alpha Scaling factor.
     * @param a The left matrix in the multiplcation operation. Not modified.
     * @param b The right matrix in the multiplcation operation. Not modified.
     * @param c Where the results of the operation are stored. Modified.
     */
    public static void mult( double alpha , DenseMatrix64F a , DenseMatrix64F b , DenseMatrix64F c )
    {
        if( b.numCols >= EjmlParameters.MULT_COLUMN_SWITCH ) {
            MatrixMatrixMult.mult_reorder(alpha,a,b,c);
        } else {
            MatrixMatrixMult.mult_small(alpha,a,b,c);
        }
    }

    /**
     * <p>Performs the following operation:<br>
     * <br>
     * c = a<sup>T</sup> * b <br>
     * <br>
     * c<sub>ij</sub> = &sum;<sub>k=1:n</sub> { a<sub>ki</sub> * b<sub>kj</sub>}
     * </p>
     *
     * @param a The left matrix in the multiplcation operation. Not modified.
     * @param b The right matrix in the multiplcation operation. Not modified.
     * @param c Where the results of the operation are stored. Modified.
     */
    public static void multTransA( DenseMatrix64F a , DenseMatrix64F b , DenseMatrix64F c )
    {
        if( b.numCols >= EjmlParameters.MULT_COLUMN_SWITCH ) {
            MatrixMatrixMult.multTransA_reorder(a,b,c);
        } else {
            MatrixMatrixMult.multTransA_small(a,b,c);
        }
    }

    /**
     * <p>Performs the following operation:<br>
     * <br>
     * c = &alpha; * a<sup>T</sup> * b <br>
     * <br>
     * c<sub>ij</sub> = &alpha; &sum;<sub>k=1:n</sub> { a<sub>ki</sub> * b<sub>kj</sub>}
     * </p>
     *
     * @param alpha Scaling factor.
     * @param a The left matrix in the multiplcation operation. Not modified.
     * @param b The right matrix in the multiplcation operation. Not modified.
     * @param c Where the results of the operation are stored. Modified.
     */
    public static void multTransA( double alpha , DenseMatrix64F a , DenseMatrix64F b , DenseMatrix64F c )
    {
        if( b.numCols >= EjmlParameters.MULT_COLUMN_SWITCH ) {
            MatrixMatrixMult.multTransA_reorder(alpha,a,b,c);
        } else {
            MatrixMatrixMult.multTransA_small(alpha,a,b,c);
        }
    }

    /**
     * <p>
     * Performs the following operation:<br>
     * <br>
     * c = a * b<sup>T</sup> <br>
     * c<sub>ij</sub> = &sum;<sub>k=1:n</sub> { a<sub>ik</sub> * b<sub>jk</sub>}
     * </p>
     *
     * @param a The left matrix in the multiplcation operation. Not modified.
     * @param b The right matrix in the multiplcation operation. Not modified.
     * @param c Where the results of the operation are stored. Modified.
     */
    public static void multTransB( DenseMatrix64F a , DenseMatrix64F b , DenseMatrix64F c )
    {
        MatrixMatrixMult.multTransB(a,b,c);
    }

    /**
     * <p>
     * Performs the following operation:<br>
     * <br>
     * c =  &alpha; * a * b<sup>T</sup> <br>
     * c<sub>ij</sub> = &alpha; &sum;<sub>k=1:n</sub> {  a<sub>ik</sub> * b<sub>jk</sub>}
     * </p>
     *
     * @param alpha Scaling factor.
     * @param a The left matrix in the multiplcation operation. Not modified.
     * @param b The right matrix in the multiplcation operation. Not modified.
     * @param c Where the results of the operation are stored. Modified.
     */
    public static void multTransB( double alpha , DenseMatrix64F a , DenseMatrix64F b , DenseMatrix64F c )
    {
        MatrixMatrixMult.multTransB(alpha,a,b,c);
    }

    /**
     * <p>
     * Performs the following operation:<br>
     * <br>
     * c = a<sup>T</sup> * b<sup>T</sup><br>
     * c<sub>ij</sub> = &sum;<sub>k=1:n</sub> { a<sub>ki</sub> * b<sub>jk</sub>}
     * </p>
     *
     * @param a The left matrix in the multiplcation operation. Not modified.
     * @param b The right matrix in the multiplcation operation. Not modified.
     * @param c Where the results of the operation are stored. Modified.
     */
    public static void multTransAB( DenseMatrix64F a , DenseMatrix64F b , DenseMatrix64F c )
    {
        if( b.numCols == 1 ) {
            // there are significantly faster algorithms when dealing with vectors
            if( a.numCols >= EjmlParameters.MULT_COLUMN_SWITCH ) {
                MatrixVectorMult.multTransA_reorder(a,b,c);
            } else {
                MatrixVectorMult.multTransA_small(a,b,c);
            }
        } else if( b.numCols >= EjmlParameters.MULT_TRANAB_COLUMN_SWITCH ) {
            MatrixMatrixMult.multTransAB_aux(a,b,c,null);
        } else {
            MatrixMatrixMult.multTransAB(a,b,c);
        }
    }

    /**
     * <p>
     * Performs the following operation:<br>
     * <br>
     * c = &alpha; * a<sup>T</sup> * b<sup>T</sup><br>
     * c<sub>ij</sub> = &alpha; &sum;<sub>k=1:n</sub> { a<sub>ki</sub> * b<sub>jk</sub>}
     * </p>
     *
     * @param alpha Scaling factor.
     * @param a The left matrix in the multiplcation operation. Not modified.
     * @param b The right matrix in the multiplcation operation. Not modified.
     * @param c Where the results of the operation are stored. Modified.
     */
    public static void multTransAB( double alpha , DenseMatrix64F a , DenseMatrix64F b , DenseMatrix64F c )
    {
        MatrixMatrixMult.multTransAB(alpha,a,b,c);
    }

    /**
     * <p>
     * Performs the following operation:<br>
     * <br>
     * c = c + a * b<br>
     * c<sub>ij</sub> = c<sub>ij</sub> + &sum;<sub>k=1:n</sub> { a<sub>ik</sub> * b<sub>kj</sub>}
     * </p>
     *
     * @param a The left matrix in the multiplcation operation. Not modified.
     * @param b The right matrix in the multiplcation operation. Not modified.
     * @param c Where the results of the operation are stored. Modified.
     */
    public static void multAdd( DenseMatrix64F a , DenseMatrix64F b , DenseMatrix64F c )
    {
        if( b.numCols >= EjmlParameters.MULT_COLUMN_SWITCH ) {
            MatrixMatrixMult.multAdd_reorder(a,b,c);
        } else {
            MatrixMatrixMult.multAdd_small(a,b,c);
        }
    }

    /**
     * <p>
     * Performs the following operation:<br>
     * <br>
     * c = c + &alpha; * a * b<br>
     * c<sub>ij</sub> = c<sub>ij</sub> +  &alpha; * &sum;<sub>k=1:n</sub> { a<sub>ik</sub> * b<sub>kj</sub>}
     * </p>
     *
     * @param alpha scaling factor.
     * @param a The left matrix in the multiplcation operation. Not modified.
     * @param b The right matrix in the multiplcation operation. Not modified.
     * @param c Where the results of the operation are stored. Modified.
     */
    public static void multAdd( double alpha , DenseMatrix64F a , DenseMatrix64F b , DenseMatrix64F c )
    {
        if( b.numCols >= EjmlParameters.MULT_COLUMN_SWITCH ) {
            MatrixMatrixMult.multAdd_reorder(alpha,a,b,c);
        } else {
            MatrixMatrixMult.multAdd_small(alpha,a,b,c);
        }
    }

    /**
     * <p>
     * Performs the following operation:<br>
     * <br>
     * c = c + a<sup>T</sup> * b<br>
     * c<sub>ij</sub> = c<sub>ij</sub> + &sum;<sub>k=1:n</sub> { a<sub>ki</sub> * b<sub>kj</sub>}
     * </p>
     *
     * @param a The left matrix in the multiplcation operation. Not modified.
     * @param b The right matrix in the multiplcation operation. Not modified.
     * @param c Where the results of the operation are stored. Modified.
     */
    public static void multAddTransA( DenseMatrix64F a , DenseMatrix64F b , DenseMatrix64F c )
    {
        if( b.numCols >= EjmlParameters.MULT_COLUMN_SWITCH ) {
            MatrixMatrixMult.multAddTransA_reorder(a,b,c);
        } else {
            MatrixMatrixMult.multAddTransA_small(a,b,c);
        }
    }

    /**
     * <p>
     * Performs the following operation:<br>
     * <br>
     * c = c + &alpha; * a<sup>T</sup> * b<br>
     * c<sub>ij</sub> =c<sub>ij</sub> +  &alpha; * &sum;<sub>k=1:n</sub> { a<sub>ki</sub> * b<sub>kj</sub>}
     * </p>
     *
     * @param alpha scaling factor
     * @param a The left matrix in the multiplcation operation. Not modified.
     * @param b The right matrix in the multiplcation operation. Not modified.
     * @param c Where the results of the operation are stored. Modified.
     */
    public static void multAddTransA( double alpha , DenseMatrix64F a , DenseMatrix64F b , DenseMatrix64F c )
    {
        if( b.numCols >= EjmlParameters.MULT_COLUMN_SWITCH ) {
            MatrixMatrixMult.multAddTransA_reorder(alpha,a,b,c);
        } else {
            MatrixMatrixMult.multAddTransA_small(alpha,a,b,c);
        }
    }

    /**
     * <p>
     * Performs the following operation:<br>
     * <br>
     * c = c + a * b<sup>T</sup> <br>
     * c<sub>ij</sub> = c<sub>ij</sub> + &sum;<sub>k=1:n</sub> { a<sub>ik</sub> * b<sub>jk</sub>}
     * </p>
     *
     * @param a The left matrix in the multiplcation operation. Not modified.
     * @param b The right matrix in the multiplcation operation. Not modified.
     * @param c Where the results of the operation are stored. Modified.
     */
    public static void multAddTransB( DenseMatrix64F a , DenseMatrix64F b , DenseMatrix64F c )
    {
        MatrixMatrixMult.multAddTransB(a,b,c);
    }

    /**
     * <p>
     * Performs the following operation:<br>
     * <br>
     * c = c + &alpha; * a * b<sup>T</sup><br>
     * c<sub>ij</sub> = c<sub>ij</sub> + &alpha; * &sum;<sub>k=1:n</sub> { a<sub>ik</sub> * b<sub>jk</sub>}
     * </p>
     *
     * @param alpha Scaling factor.
     * @param a The left matrix in the multiplcation operation. Not modified.
     * @param b The right matrix in the multiplcation operation. Not modified.
     * @param c Where the results of the operation are stored. Modified.
     */
    public static void multAddTransB( double alpha , DenseMatrix64F a , DenseMatrix64F b , DenseMatrix64F c )
    {
        MatrixMatrixMult.multAddTransB(alpha,a,b,c);
    }

    /**
     * <p>
     * Performs the following operation:<br>
     * <br>
     * c = c + a<sup>T</sup> * b<sup>T</sup><br>
     * c<sub>ij</sub> = c<sub>ij</sub> + &sum;<sub>k=1:n</sub> { a<sub>ki</sub> * b<sub>jk</sub>}
     * </p>
     *
     * @param a The left matrix in the multiplcation operation. Not Modifed.
     * @param b The right matrix in the multiplcation operation. Not Modifed.
     * @param c Where the results of the operation are stored. Modified.
     */
    public static void multAddTransAB( DenseMatrix64F a , DenseMatrix64F b , DenseMatrix64F c )
    {
        if( b.numRows == 1 ) {
            // there are significantly faster algorithms when dealing with vectors
            if( a.numCols >= EjmlParameters.MULT_COLUMN_SWITCH ) {
                MatrixVectorMult.multAddTransA_reorder(a,b,c);
            } else {
                MatrixVectorMult.multAddTransA_small(a,b,c);
            }
        } else if( b.numCols >= EjmlParameters.MULT_TRANAB_COLUMN_SWITCH ) {
            MatrixMatrixMult.multAddTransAB_aux(a,b,c,null);
        } else {
            MatrixMatrixMult.multAddTransAB(a,b,c);
        }
    }

    /**
     * <p>
     * Performs the following operation:<br>
     * <br>
     * c = c + &alpha; * a<sup>T</sup> * b<sup>T</sup><br>
     * c<sub>ij</sub> = c<sub>ij</sub> + &alpha; * &sum;<sub>k=1:n</sub> { a<sub>ki</sub> * b<sub>jk</sub>}
     * </p>
     *
     * @param alpha Scaling factor.
     * @param a The left matrix in the multiplcation operation. Not Modifed.
     * @param b The right matrix in the multiplcation operation. Not Modifed.
     * @param c Where the results of the operation are stored. Modified.
     */
    public static void multAddTransAB( double alpha , DenseMatrix64F a , DenseMatrix64F b , DenseMatrix64F c )
    {
        if( b.numCols >= EjmlParameters.MULT_TRANAB_COLUMN_SWITCH ) {
            MatrixMatrixMult.multAddTransAB_aux(alpha,a,b,c,null);
        } else {
            MatrixMatrixMult.multAddTransAB(alpha,a,b,c);
        }
    }

    /**
     * <p>Performs the an element by element multiplcation operation:<br>
     * <br>
     * a<sub>ij</sub> = a<sub>ij</sub> * b<sub>ij</sub> <br>
     * </p>
     * @param a The left matrix in the multiplcation operation. Modified.
     * @param b The right matrix in the multiplcation operation. Not modified.
     */
    public static void multElement( DenseMatrix64F a , DenseMatrix64F b )
    {
        if( a.numCols != b.numCols || a.numRows != b.numRows ) {
            throw new RuntimeException("The 'a' and 'b' matrices do not have compatable dimensions");
        }

        int length = a.getNumElements();
        double dataA[] = a.data;
        double dataB[] = b.data;

        for( int i = 0; i < length; i++ ) {
            dataA[i] *= dataB[i];
        }
    }

    /**
     * <p>Performs the an element by element multiplcation operation:<br>
     * <br>
     * c<sub>ij</sub> = a<sub>ij</sub> * b<sub>ij</sub> <br>
     * </p>
     * @param a The left matrix in the multiplcation operation. Not modified.
     * @param b The right matrix in the multiplcation operation. Not modified.
     * @param c Where the results of the operation are stored. Modified.
     */
    public static void multElement( DenseMatrix64F a , DenseMatrix64F b , DenseMatrix64F c )
    {
        if( a.numCols != b.numCols || a.numRows != b.numRows
                || a.numRows != c.numRows || a.numCols != c.numCols ) {
            throw new RuntimeException("The 'a' and 'b' matrices do not have compatable dimensions");
        }

        int length = a.getNumElements();
        double dataA[] = a.data;
        double dataB[] = b.data;
        double dataC[] = c.data;

        for( int i = 0; i < length; i++ ) {
            dataC[i] = dataA[i] * dataB[i];
        }
    }

    /**
     * <p>Performs the following operation:<br>
     * <br>
     * a = a + b <br>
     * a<sub>ij</sub> = a<sub>ij</sub> + b<sub>ij</sub> <br>
     * </p>
     *
     * @param a A Matrix. Modified.
     * @param b A Matrix. Not modified.
     */
    public static void addEquals( DenseMatrix64F a , DenseMatrix64F b )
    {
        if( a.numCols != b.numCols || a.numRows != b.numRows ) {
            throw new RuntimeException("The 'a' and 'b' matrices do not have compatable dimensions");
        }

        final double dataA[] = a.data;
        final double dataB[] = b.data;

        final int length = a.getNumElements();

        for( int i = 0; i < length; i++ ) {
            dataA[i] += dataB[i];
        }
    }

    /**
     * <p>Performs the following operation:<br>
     * <br>
     * a = a +  &beta; * b  <br>
     * a<sub>ij</sub> = a<sub>ij</sub> + &beta; * b<sub>ij</sub>
     * </p>
     *
     * @param beta The number that matrix 'b' is multiplied by.
     * @param a A Matrix. Modified.
     * @param b A Matrix. Not modified.
     */
    public static void addEquals( DenseMatrix64F a , double beta, DenseMatrix64F b )
    {
        if( a.numCols != b.numCols || a.numRows != b.numRows ) {
            throw new RuntimeException("The 'a' and 'b' matrices do not have compatable dimensions");
        }

        final double dataA[] = a.data;
        final double dataB[] = b.data;

        final int length = a.getNumElements();

        for( int i = 0; i < length; i++ ) {
            dataA[i] += beta * dataB[i];
        }
    }

    /**
     * <p>Performs the following operation:<br>
     * <br>
     * c = a + b <br>
     * c<sub>ij</sub> = a<sub>ij</sub> + b<sub>ij</sub> <br>
     * </p>
     *
     * <p>
     * Matrix C can be the same instance as Matrix A and/or B.
     * </p>
     *
     * @param a A Matrix. Not modified.
     * @param b A Matrix. Not modified.
     * @param c A Matrix where the results are stored. Modified.
     */
    public static void add( DenseMatrix64F a , DenseMatrix64F b , DenseMatrix64F c )
    {
        if( a.numCols != b.numCols || a.numRows != b.numRows
                || a.numCols != c.numCols || a.numRows != c.numRows ) {
            throw new RuntimeException("The matrices are not all the same dimension.");
        }

        final double dataA[] = a.data;
        final double dataB[] = b.data;
        final double dataC[] = c.data;

        final int length = a.getNumElements();

        for( int i = 0; i < length; i++ ) {
            dataC[i] = dataA[i]+dataB[i];
        }
    }

    /**
     * <p>Performs the following operation:<br>
     * <br>
     * c = a + &beta; * b <br>
     * c<sub>ij</sub> = a<sub>ij</sub> + &beta; * b<sub>ij</sub> <br>
     * </p>
     *
     * <p>
     * Matrix C can be the same instance as Matrix A and/or B.
     * </p>
     *
     * @param a A Matrix. Not modified.
     * @param beta Scaling factor for matrix b.
     * @param b A Matrix. Not modified.
     * @param c A Matrix where the results are stored. Modified.
     */
    public static void add( DenseMatrix64F a , double beta , DenseMatrix64F b , DenseMatrix64F c )
    {
        if( a.numCols != b.numCols || a.numRows != b.numRows
                || a.numCols != c.numCols || a.numRows != c.numRows ) {
            throw new RuntimeException("The matrices are not all the same dimension.");
        }

        final double dataA[] = a.data;
        final double dataB[] = b.data;
        final double dataC[] = c.data;

        final int length = a.getNumElements();

        for( int i = 0; i < length; i++ ) {
            dataC[i] = dataA[i]+beta*dataB[i];
        }
    }

    /**
     * <p>Performs the following operation:<br>
     * <br>
     * c = &alpha; * a + &beta; * b <br>
     * c<sub>ij</sub> = &alpha; * a<sub>ij</sub> + &beta; * b<sub>ij</sub> <br>
     * </p>
     *
     * <p>
     * Matrix C can be the same instance as Matrix A and/or B.
     * </p>
     *
     * @param alpha A scaling factor for matrix a.
     * @param a A Matrix. Not modified.
     * @param beta A scaling factor for matrix b.
     * @param b A Matrix. Not modified.
     * @param c A Matrix where the results are stored. Modified.
     */
    public static void add( double alpha , DenseMatrix64F a , double beta , DenseMatrix64F b , DenseMatrix64F c )
    {
        if( a.numCols != b.numCols || a.numRows != b.numRows
                || a.numCols != c.numCols || a.numRows != c.numRows ) {
            throw new RuntimeException("The matrices are not all the same dimension.");
        }

        final double dataA[] = a.data;
        final double dataB[] = b.data;
        final double dataC[] = c.data;

        final int length = a.getNumElements();

        for( int i = 0; i < length; i++ ) {
            dataC[i] = alpha*dataA[i]+beta*dataB[i];
        }
    }

    /**
     * <p>Performs the following operation:<br>
     * <br>
     * c = &alpha; * a + b <br>
     * c<sub>ij</sub> = &alpha; * a<sub>ij</sub> + b<sub>ij</sub> <br>
     * </p>
     *
     * <p>
     * Matrix C can be the same instance as Matrix A and/or B.
     * </p>
     *
     * @param alpha A scaling factor for matrix a.
     * @param a A Matrix. Not modified.
     * @param b A Matrix. Not modified.
     * @param c A Matrix where the results are stored. Modified.
     */
    public static void add( double alpha , DenseMatrix64F a , DenseMatrix64F b , DenseMatrix64F c )
    {
        if( a.numCols != b.numCols || a.numRows != b.numRows
                || a.numCols != c.numCols || a.numRows != c.numRows ) {
            throw new RuntimeException("The matrices are not all the same dimension.");
        }

        final double dataA[] = a.data;
        final double dataB[] = b.data;
        final double dataC[] = c.data;

        final int length = a.getNumElements();

        for( int i = 0; i < length; i++ ) {
            dataC[i] = alpha*dataA[i]+dataB[i];
        }
    }

    /**
     * <p>Performs an in-place scalar addition:<br>
     * <br>
     * a = a + val<br>
     * a<sub>ij</sub> = a<sub>ij</sub> + val<br>
     * </p>
     *
     * @param a A matrix.  Modified.
     * @param val The value that's added to each element.
     */
    public static void add( DenseMatrix64F a , double val ) {
        final double dataA[] = a.data;

        final int length = a.getNumElements();

        for( int i = 0; i < length; i++ ) {
            dataA[i] += val;
        }
    }

    /**
     * <p>Performs scalar addition:<br>
     * <br>
     * c = a + val<br>
     * c<sub>ij</sub> = a<sub>ij</sub> + val<br>
     * </p>
     *
     * @param a A matrix. Not modified.
     * @param c A matrix. Modified.
     * @param val The value that's added to each element.
     */
    public static void add( DenseMatrix64F a , double val , DenseMatrix64F c ) {
        if( a.numRows != c.numRows || a.numCols != c.numCols ) {
            throw new IllegalArgumentException("Dimensions of a and c do not match.");
        }
        final double dataA[] = a.data;
        final double dataC[] = c.data;

        final int length = a.getNumElements();

        for( int i = 0; i < length; i++ ) {
            dataC[i] = dataA[i]+val;
        }
    }

    /**
     * <p>Performs the following subtraction operation:<br>
     * <br>
     * a = a - b  <br>
     * a<sub>ij</sub> = a<sub>ij</sub> - b<sub>ij</sub>
     * </p>
     *
     * @param a A Matrix. Modified.
     * @param b A Matrix. Not modified.
     */
    public static void subEquals( DenseMatrix64F a , DenseMatrix64F b )
    {
        if( a.numCols != b.numCols || a.numRows != b.numRows ) {
            throw new RuntimeException("The 'a' and 'b' matrices do not have compatable dimensions");
        }

        final double dataA[] = a.data;
        final double dataB[] = b.data;

        final int length = a.getNumElements();

        for( int i = 0; i < length; i++ ) {
            dataA[i] -= dataB[i];
        }
    }

    /**
     * <p>Performs the following subtraction operation:<br>
     * <br>
     * c = a - b  <br>
     * c<sub>ij</sub> = a<sub>ij</sub> - b<sub>ij</sub>
     * </p>
     * <p>
     * Matrix C can be the same instance as Matrix A and/or B.
     * </p>
     *
     * @param a A Matrix. Not modified.
     * @param b A Matrix. Not modified.
     * @param c A Matrix. Modified.
     */
    public static void sub( DenseMatrix64F a , DenseMatrix64F b , DenseMatrix64F c )
    {
        if( a.numCols != b.numCols || a.numRows != b.numRows ) {
            throw new RuntimeException("The 'a' and 'b' matrices do not have compatable dimensions");
        }

        final double dataA[] = a.data;
        final double dataB[] = b.data;
        final double dataC[] = c.data;

        final int length = a.getNumElements();

        for( int i = 0; i < length; i++ ) {
            dataC[i] = dataA[i] - dataB[i];
        }
    }

    /**
     * <p>
     * Solves for x:<br>
     * <br>
     * x = a<sup>-1</sup>B<br>
     * </p>
     *
     * <p>
     * If the system could not be solved then false is returned.  If it returns true
     * that just means the algorithm finished operating, but the results could still be bad
     * because 'A' is singular or nearly singular.
     * </p>
     *
     * <p>
     * It is ok for 'b' and 'x' to be the same matrix.
     * </p>
     *
     * @param a A matrix that is m by m. Not modified.
     * @param b A matrix that is m by n. Not modified.
     * @param x A matrix that is m by n. Modified.
     *
     * @return true if it could invert the matrix false if it could not.
     */
    public static boolean solve( DenseMatrix64F a , DenseMatrix64F b , DenseMatrix64F x )
    {
        LinearSolver solver;

        if( a.numRows == a.numCols ) {
            LUDecompositionAlt alg = new LUDecompositionAlt();

            solver = new LinearSolverLu(alg);
        } else if( a.numRows > a.numCols ){
            solver = new LinearSolverQrHouseCol();
        } else {
            throw new IllegalArgumentException("Can't solve since there is an infinite number of solutions.");
        }

        if( !solver.setA(a) )
            return false;

        solver.solve(b,x);
        return true;
    }

    /**
     * <p>
     * Performs an in-place element by element scalar multiplation.<br>
     * <br>
     * a<sub>ij</sub> = &alpha;*a<sub>ij</sub>
     * </p>
     *
     * @param a The matrix that is to be scalled.  Modified.
     * @param alpha the amount each element is multiplied by.
     */
    public static void scale( double alpha , DenseMatrix64F a )
    {
        final double data[] = a.data;
        final int size = a.getNumElements();

        for( int i = 0; i < size; i++ ) {
            data[i] *= alpha;
        }
    }

    /**
     * <p>
     * Performs an element by element scalar multiplation.<br>
     * <br>
     * b<sub>ij</sub> = &alpha;*a<sub>ij</sub>
     * </p>
     *
     * @param a The matrix that is to be scalled.  Modified.
     * @param alpha the amount each element is multiplied by.
     */
    public static void scale( double alpha , DenseMatrix64F a , DenseMatrix64F b)
    {
        if( a.numRows != b.numRows || a.numCols != b.numCols )
            throw new IllegalArgumentException("Matrices must have the same shape");

        final double dataA[] = a.data;
        final double dataB[] = b.data;
        final int size = a.getNumElements();

        for( int i = 0; i < size; i++ ) {
            dataB[i] = dataA[i]*alpha;
        }
    }

    /**
     * <p>
     * Sets every element in the matrix to the specified value.<br>
     * <br>
     * a<sub>ij</sub> = value
     * <p>
     *
     * @param a A matrix whose elements are about to be set. Modified.
     * @param value The value each element will have.
     */
    public static void set( DenseMatrix64F a , double value )
    {
        final double data[] = a.data;
        final int size = a.getNumElements();

        for( int i = 0; i < size; i++ ) {
            data[i] = value;
        }
    }

    /**
     * Performs an in-place transpose.  This algorithm is only efficient for square
     * matrices.
     *
     * @param mat The matrix that is to be transposed. Modified.
     */
    public static void transpose( DenseMatrix64F mat ) {
        if( mat.numCols == mat.numRows ){
            final double data[] = mat.data;

            for( int i = 0; i < mat.numRows; i++ ) {
                int index = i*mat.numCols+i+1;
                for( int j = i+1; j < mat.numCols; j++ ) {
                    int otherIndex = j*mat.numCols+i;
                    double val = data[index];
                    data[index] = data[otherIndex];
                    data[otherIndex] = val;
                    index++;
                }
            }
        } else {
            DenseMatrix64F b = new DenseMatrix64F(mat.numCols,mat.numRows);
            transpose(mat,b);
            mat.setReshape(b);
        }
    }

    /**
     * <p>
     * Transposes matrix 'a' and stores the results in 'b':<br>
     * <br>
     * b<sub>ij</sub> = a<sub>ji</sub><br>
     * where 'b' is the transpose of 'a'.
     * </p>
     *
     * @param a The original matrix.  Not modified.
     * @param b Where the transpose is stored. Modified.
     */
    public static void transpose( DenseMatrix64F a , DenseMatrix64F b )
    {
        if( a.numRows != b.numCols || a.numCols != b.numRows ) {
            throw new RuntimeException("Incompatible matrix dimensions");
        }
        final double rdata[] = b.data;
        final double data[] = a.data;

        int index = 0;
        for( int i = 0; i < b.numRows; i++ ) {
            int index2 = i;

            for( int j = 0; j < b.numCols; j++ ) {
                rdata[index++] = data[index2];
                index2 += a.numCols;
            }
        }
    }

    /**
     * <p>
     * This computes the Frobenius norm of the matrix:<br>
     * <br>
     * normF = Sqrt{  &sum;<sub>i=1:m</sub> &sum;<sub>j=1:n</sub> { a<sub>ij</sub><sup>2</sup>}   }
     * </p>
     * <p>
     * This implementation of the Frobenius norm is a straight forward implementation and can
     * be suspectable for overflow/underflow issues.  A more resilient implementation is
     * {@link #normF_alt}.
     * </p>
     *
     * @param a The matrix whose norm is computed.  Not modified.
     */
    public static double normF( DenseMatrix64F a ) {
        double total = 0;

        final double data[] = a.data;

        for( int i = 0; i < data.length; i++ ) {
            double val = data[i];
            total += val*val;
        }

        return Math.sqrt(total);
    }

    /**
     * <p>
     * This computes the Frobenius norm of the matrix in a way that attempts to reduce the chance
     * of an overflow/underflow:<br>
     * <br>
     * maxVal = Max{|a<sub>ij</sub>|}<br>
     * normF = maxVal * Sqrt{ &sum;<sub>i=1:m</sub> &sum;<sub>j=1:n</sub> { a<sub>ij</sub><sup>2</sup>/maxVal<sup>2</sup> }  }
     * </p>
     * <p>
     * The elements are first normalized by the {@link #maxAbs} of the matrix and then the results are multiplied by the same number.
     * </p>
     *
     * @param a The matrix whose norm is computed.  Not modified.
     */
    public static double normF_alt( DenseMatrix64F a ) {
        double total = 0;

        double scale = maxAbs(a);

        final double data[] = a.data;

        for( int i = 0; i < data.length; i++ ) {
            double val = data[i]/scale;
            total += val*val;
        }

        return scale*Math.sqrt(total);
    }

    /**
     * <p>
     * Returns the value of the element in the matrix that has the largest value.<br>
     * <br>
     * Max{ a<sub>ij</sub> } for all i and j<br>
     * </p>
     *
     * @param a A matrix.
     * @return The max element value of the matrix.
     */
    public static double max( DenseMatrix64F a ) {
        final int size = a.getNumElements();

        final double dataA[] = a.data;

        double max = dataA[0];
        for( int i = 1; i < size; i++ ) {
            double val = dataA[i];
            if( val >= max ) {
                max = val;
            }
        }

        return max;
    }

    /**
     * <p>
     * Returns the absolute value of the element in the matrix that has the largest absolute value.<br>
     * <br>
     * Max{ |a<sub>ij</sub>| } for all i and j<br>
     * </p>
     *
     * @param a A matrix.
     * @return The max element value of the matrix.
     */
    public static double maxAbs( DenseMatrix64F a ) {
        final int size = a.getNumElements();

        final double dataA[] = a.data;

        double max = 0;
        for( int i = 0; i < size; i++ ) {
            double val = Math.abs(dataA[i]);
            if( val >= max ) {
                max = val;
            }
        }

        return max;
    }

    /**
     * <p>
     * Returns the value of the element in the matrix that has the minimum value.<br>
     * <br>
     * Min{ a<sub>ij</sub> } for all i and j<br>
     * </p>
     *
     * @param a A matrix.
     * @return The value of element in the matrix with the minimum value.
     */
    public static double min( DenseMatrix64F a ) {
        final int size = a.getNumElements();

        final double dataA[] = a.data;

        double min = dataA[0];
        for( int i = 1; i < size; i++ ) {
            double val = dataA[i];
            if( val < min ) {
                min = val;
            }
        }

        return min;
    }

    /**
     * <p>
     * This computes the trace of the matrix:<br>
     * <br>
     * trace = &sum;<sub>i=1:n</sub> { a<sub>ii</sub> }
     * </p>
     * <p>
     * The trace is only defined for square matrices.
     * </p>
     *
     * @param a A square matrix.  Not modified.
     */
    public static double trace( DenseMatrix64F a ) {
        if( a.numRows != a.numCols ) {
            throw new IllegalArgumentException("The matrix must be square");
        }

        final double data[] = a.data;

        double sum = 0;
        int index = 0;
        for( int i = 0; i < a.numRows; i++ ) {
            sum += data[index];
            index += 1 + a.numCols;
        }

        return sum;
    }

    /**
     * Returns the determinant of the matrix.  If the inverse of the matrix is also
     * needed, then using {@link org.ejml.alg.dense.decomposition.lu.LUDecompositionAlt} directly (or any
     * similar algorithm) can be more efficient.
     *
     * @param mat The matrix whose determinant is to be computed.  Not modified.
     * @return The determinant.
     */
    public static double det( DenseMatrix64F mat )
    {

        int numCol = mat.getNumCols();
        int numRow = mat.getNumRows();

        if( numCol != numRow ) {
            throw new IllegalArgumentException("Must be a square matrix.");
        } else if( numCol <= 4 ) {
            // slight performance boost overall by doing it this way
            // when it was the case statement the VM did some strange optimization
            // and made case 2 about 1/2 the speed
            if( numCol == 2 ) {
                return UtilDeterminant.det2by2(mat.data);
            } else if( numCol == 3 ) {
                return UtilDeterminant.det3by3(mat.data);
            } else if( numCol == 4 ) {
                return UtilDeterminant.det4by4(mat.data);
            } else if( numCol == 1 ) {
                return mat.data[0];
            } else {
                 throw new RuntimeException("Egads");
            }
        } else {
            LUDecompositionAlt alg = new LUDecompositionAlt();
            if( !alg.decompose(mat) )
                return 0.0;
            return alg.computeDeterminant();
        }
    }

    /**
     * <p>
     * Performs a matrix inversion operation on the specified matrix and stores the results
     * in the same matrix.<br>
     * <br>
     * a = a<sup>-1<sup>
     * </p>
     *
     * <p>
     * If the algorithm could not invert the matrix then false is returned.  If it returns true
     * that just means the algorithm finished.  The results could still be bad
     * because the matrix is singular or nearly singular.
     * </p>
     *
     * @param mat The matrix that is to be inverted.  Results are stored here.  Modified.
     * @return true if it could invert the matrix false if it could not.
     */
    public static boolean invert( DenseMatrix64F mat) {
        if( mat.numCols <= 3 ) {
            if( mat.numCols != mat.numRows ) {
                throw new IllegalArgumentException("Must be a square matrix.");
            }

            if( mat.numCols == 2 ) {
                MatrixInvertSpecialized.invert2x2(mat.data);
            } else if( mat.numCols == 3 ) {
                MatrixInvertSpecialized.invert3x3(mat.data);
            } else {
                mat.data[0] = 1.0/mat.data[0];
            }
        } else {
            LUDecompositionAlt alg = new LUDecompositionAlt();
            LinearSolverLu solver = new LinearSolverLu(alg);
            if( solver.setA(mat) ) {
                solver.invert(mat);
            } else {
                return false;
            }
        }
        return true;
    }

    /**
     * <p>
     * Performs a matrix inversion operation that does not modify the original
     * and stores the results in another matrix.  The two matrices must have the
     * same dimension.<br>
     * <br>
     * b = a<sup>-1<sup>
     * </p>
     *
     * <p>
     * If the algorithm could not invert the matrix then false is returned.  If it returns true
     * that just means the algorithm finished.  The results could still be bad
     * because the matrix is singular or nearly singular.
     * </p>
     *
     * @param mat The matrix that is to be inverted. Not modified.
     * @param result Where the inverse matrix is stored.  Modified.
     * @return true if it could invert the matrix false if it could not.
     */
    public static boolean invert( DenseMatrix64F mat, DenseMatrix64F result ) {
        if( mat.numCols <= 3 ) {
            if( mat.numCols != mat.numRows ) {
                throw new IllegalArgumentException("Must be a square matrix.");
            }
            result.set(mat);

            if( result.numCols == 2 ) {
                MatrixInvertSpecialized.invert2x2(result.data);
            } else if( result.numCols == 3 ) {
                MatrixInvertSpecialized.invert3x3(result.data);
            } else {
                result.data[0] = 1.0/result.data[0];
            }
        } else {
            LUDecompositionAlt alg = new LUDecompositionAlt();
            LinearSolverLu solver = new LinearSolverLu(alg);

            if( !solver.setA(mat))
                return false;
            solver.invert(result);
        }
        return true;
    }
}
