/*
 * Copyright (c) 2009, Peter Abeles. All Rights Reserved.
 *
 * This file is part of Efficient Java Matrix Library (EJML).
 *
 * EJML is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 *
 * EJML is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with EJML.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.ejml.ops;

import org.ejml.alg.dense.decomposition.chol.CholeskyDecompositionBasic;
import org.ejml.data.DenseMatrix64F;


/**
 * Used to compute features that describe the structure of a matrix.
 *
 * @author Peter Abeles
 */
public class MatrixFeatures {

    /**
     * Checks to see if any element in the matrix is NaN.
     *
     * @param m A matrix. Not modified.
     * @return True if any element in the matrix is NaN.
     */
    public static boolean hasNaN( DenseMatrix64F m )
    {
        int length = m.getNumElements();
        double data[] = m.data;

        for( int i = 0; i < length; i++ ) {
            if( Double.isNaN(data[i]))
                return true;
        }
        return false;
    }

    /**
     * Checks to see if any element in the matrix is NaN of Infinite.
     *
     * @param m A matrix. Not modified.
     * @return True if any element in the matrix is NaN of Infinite.
     */
    public static boolean hasUncountable( DenseMatrix64F m )
    {
        int length = m.getNumElements();
        double data[] = m.data;

        for( int i = 0; i < length; i++ ) {
            double a = data[i];
            if( Double.isNaN(a) || Double.isInfinite(a))
                return true;
        }
        return false;
    }

    /**
     * Checks to see if the matrix is a vector or not.
     *
     * @param mat A matrix. Not modified.
     *
     * @return True if it is a vector and false if it is not.
     */
    public static boolean isVector( DenseMatrix64F mat ) {
        return (mat.numCols == 1 || mat.numRows == 1);
    }

    /**
     * <p>
     * Checks to see if the matrix is positive definite.
     * </p>
     * <p>
     * x<sup>T</sup> A x > 0<br>
     * for all x where x is a non-zero vector and A is a symmetric matrix.
     * </p>
     * <p>
     * *Note* this class does NOT check to see if the input is symmetric.
     * </p>
     *
     *
     * @param A square symmetric matrix. Not modified.
     *
     * @return True if it is positive definite and false if it is not.
     */
    public static boolean isPositiveDefinite( DenseMatrix64F A ) {

        if( A.numRows <= 4 ) {
            return( CommonOps.det(A) < 0 );
        }

        CholeskyDecompositionBasic chol = new CholeskyDecompositionBasic(false,true);
        return chol.decompose(A);
    }

    /**
     * <p>
     * Checks to see if the matrix is positive semidefinite.
     * </p>
     * <p>
     * x<sup>T</sup> A x >= 0<br>
     * for all x where x is a non-zero vector and A is a symmetric matrix.
     * </p>
     * <p>
     * *Note* this class does NOT check to see if the input is symmetric.
     * </p>
     *
     * @param A square symmetric matrix. Not modified.
     *
     * @return True if it is positive semidefinite and false if it is not.
     */
    public static boolean isPositiveSemidefinite( DenseMatrix64F A ) {
        if( A.numRows != A.numCols )
            throw new IllegalArgumentException("A is not square");

        // a matrix can be tested for positive definite by seeing if specific submatrices have
        // a determinant that is strictly positive.
        // The submatrices are formed by removing the outter most rows and columns.
        if( CommonOps.det(A) < 0 )
            return false;

        DenseMatrix64F temp = new DenseMatrix64F(A.numRows-1,A.numCols-1);

        for( int i = 1; i < A.numRows-1; i++ ) {
            SpecializedOps.submatrix(A,0,i,0,i,temp);
            if(CommonOps.det(temp) < 0 )
                return false;
        }

       return true;
    }

    /**
     * Checks to see if it is a square matrix.  A square matrix has
     * the same number of rows and columns.
     *
     * @param mat A matrix. Not modified.
     * @return True if it is a square matrix and false if it is not.
     */
    public static boolean isSquare( DenseMatrix64F mat ) {
        return mat.numCols == mat.numRows;
    }

    /**
     * <p>
     * Returns true if the matrix is symmetric within the tolerance.  Only square matrices can be symetric.
     * </p>
     * <p>
     * A matrix is symetric if:<br>
     * |a<sub>ij</sub> - a<sub>ji</sub>| &le; tol
     * </p>
     *
     * @param m A matrix. Not modified.
     * @param tol Tolerance for how similar two elements need to be.
     * @return true if it is symmetric and false if it is not.
     */
    public static boolean isSymmetric( DenseMatrix64F m , double tol ) {
        if( m.numCols != m.numRows )
            return false;

        for( int i = 0; i < m.numRows; i++ ) {
            for( int j = 0; j < i; j++ ) {
                double a = m.get(i,j);
                double b = m.get(j,i);

                double diff = Math.abs(a-b);

                if( diff > tol ) {
                    return false;
                }
            }
        }
        return true;
    }

    /**
     * <p>
     * Returns true if the matrix is perfectly symetric.  Only square matrices can be symetric.
     * </p>
     * <p>
     * A matrix is symetric if:<br>
     * a<sub>ij</sub> == a<sub>ji</sub>
     * </p>
     *
     * @param m A matrix. Not modified.
     * @return true if it is symetric and false if it is not.
     */
    public static boolean isSymmetric( DenseMatrix64F m ) {
        return isSymmetric(m,0.0);
    }

    /**
     * <p>
     * Checks to see if each element in the two matrices are within tolerance of
     * each other.
     * <p>
     *
     * <p>
     * The two matrices are similar if:<br>
     * |a<sub>ij</sub> - b<sub>ij</sub>| &le; tol
     * </p>
     *
     * @param a A matrix. Not modified.
     * @param b A matrix. Not modified.
     * @param tol How similiar each element needs to be.
     * @return true if similiar and false otherwise.
     */
    public static boolean isSimilar( DenseMatrix64F a , DenseMatrix64F b , double tol )
    {
        if( a.numRows != b.numRows || a.numCols != b.numCols ) {
            return false;
        }

        final double dataA[] = a.data;
        final double dataB[] = b.data;
        final int length = a.getNumElements();

        for( int i = 0; i < length; i++ ) {
            double diff = dataA[i] - dataB[i];
            if( diff < 0 ) diff = -diff;

            if( diff > tol ) {
                return false;
            }
        }
        return true;
    }

    /**
     * <p>
     * Checks to see if each element in the two matrices are identical.
     * <p>
     *
     * <p>
     * The two matrices are identical if:<br>
     * a<sub>ij</sub> == b<sub>ij</sub>
     * </p>
     * 
     * @param a A matrix. Not modified.
     * @param b A matrix. Not modified.
     * @return true if identical and false otherwise.
     */
    public static boolean isIdentical( DenseMatrix64F a, DenseMatrix64F b ) {
        if( a.numRows != b.numRows || a.numCols != b.numCols ) {
            return false;
        }

        int length = a.getNumElements();
        for( int i = 0; i < length; i++ ) {
            if( a.get(i) != b.get(i)) {
                return false;
            }
        }

        return true;
    }

    public static boolean isOrthogonal( DenseMatrix64F Q , double tol )
    {
        if( Q.numCols != Q.numRows )
            return false;

        DenseMatrix64F QQT = new DenseMatrix64F(Q.numCols,Q.numCols);

        CommonOps.multTransB(Q,Q,QQT);

        return isIdentity(QQT,tol);
    }

    public static boolean isIdentity( DenseMatrix64F mat , double tol )
    {
        // see if the result is an identity matrix
        int index = 0;
        for( int i = 0; i < mat.numRows; i++ ) {
            for( int j = 0; j < mat.numCols; j++ ) {
                if( i == j ) {
                    if( Math.abs(mat.data[index++]-1) > tol )
                        return false;
                } else {
                    if( Math.abs(mat.data[index++]) > tol )
                        return false;
                }
            }
        }

        return true;
    }

    /**
     * Checks to see if all the diagonal elements in the matrix are positive.
     *
     * @param a A matrix. Not modified.
     * @return true if all the  diagonal elements are positive, false otherwise.
     */
    public static boolean isDiagonalPositive( DenseMatrix64F a ) {
        for( int i = 0; i < a.numRows; i++ ) {
            if( a.get(i,i) < 0 )
                return false;
        }
        return true;
    }

    // TODO write this
    public static boolean isFullRank( DenseMatrix64F a ) {
        throw new RuntimeException("Implement");
    }
}
