/*
 * Copyright (c) 2009, Peter Abeles. All Rights Reserved.
 *
 * This file is part of Efficient Java Matrix Library (EJML).
 *
 * EJML is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 *
 * EJML is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with EJML.  If not, see <http://www.gnu.org/licenses/>.
 */

package org.ejml.ops;

import org.ejml.alg.dense.decomposition.*;
import org.ejml.alg.dense.decomposition.lu.LUDecompositionAlt;
import org.ejml.alg.dense.decomposition.svd.SvdNumericalRecipes;
import org.ejml.data.DenseMatrix64F;


/**
 * This contains less common or more specialized matrix operations.
 *
 * @author Peter Abeles
 */
// TODO change decompositions so that a new decompose is created instead of returning one which
    // contains the decomposition to encourage reuse?
public class SpecializedOps {

    /**
     * Creates a copy of a matrix but swaps the rows as specified by the order array.
     *
     * @param order Specifies which row in the dest corresponds to a row in the src. Not modified.
     * @param src The original matrix. Not modified.
     * @param dst A Matrix that is a row swapped copy of src. Modified.
     */
    public static void copyChangeRow( int order[] , DenseMatrix64F src , DenseMatrix64F dst )
    {
        double dataSrc[] = src.data;
        double dataDst[] = dst.data;

        for( int i = 0; i < src.numRows; i++ ) {
            int indexDst = i*src.numCols;
            int indexSrc = order[i]*src.numCols;

            System.arraycopy(dataSrc,indexSrc,dataDst,indexDst,src.numCols);
        }
    }

    /**
     * <p>
     * Computes the F norm of the difference between the two Matrices:<br>
     * <br>
     * Sqrt{&sum;<sub>i=1:m</sub> &sum;<sub>j=1:n</sub> ( a<sub>ij</sub> - b<sub>ij</sub>)<sup>2</sup>}
     * </p>
     * <p>
     * This is often used as a cost function.
     * </p>
     *
     * @see CommonOps#normF
     *
     * @param a m by n matrix. Not modified.
     * @param b m by n matrix. Not modified.
     *
     * @return The F normal of the difference matrix.
     */
    public static double diffNormF( DenseMatrix64F a , DenseMatrix64F b )
    {
        if( a.numRows != b.numRows || a.numCols != b.numCols ) {
            throw new IllegalArgumentException("Both matrices must have the same shape.");
        }

        final double dataA[] = a.data;
        final double dataB[] = b.data;

        final int size = a.getNumElements();

        double total=0;
        for( int i = 0; i < size; i++ ) {
            double diff = dataB[i] - dataA[i];
            total += diff*diff;
        }
        return Math.sqrt(total);
    }

    /**
     * <p>
     * Computes the p=1 p-norm of the difference between the two Matrices:<br>
     * <br>
     * &sum;<sub>i=1:m</sub> &sum;<sub>j=1:n</sub> | a<sub>ij</sub> - b<sub>ij</sub>| <br>
     * <br>
     * where |x| is the absolute value of x.
     * </p>
     * <p>
     * This is often used as a cost function.
     * </p>
     *
     * @param a m by n matrix. Not modified.
     * @param b m by n matrix. Not modified.
     *
     * @return The p=1 p-norm of the difference matrix.
     */
    public static double diffNormP1( DenseMatrix64F a , DenseMatrix64F b )
    {
        if( a.numRows != b.numRows || a.numCols != b.numCols ) {
            throw new IllegalArgumentException("Both matrices must have the same shape.");
        }

        final double dataA[] = a.data;
        final double dataB[] = b.data;

        final int size = a.getNumElements();

        double total=0;
        for( int i = 0; i < size; i++ ) {
            total += Math.abs(dataB[i] - dataA[i]);
        }
        return total;
    }

    /**
     * <p>
     * This computes the infinity norm of the matrix:<br>
     * <br>
     * normInf = max{ i=1:m j=1:n ,  |a<sub>ij</sub>| }
     * </p>
     * <p>
     * In other words this norm is the largest absolute value of all the elements.
     * </p>
     *
     * @param a The matrix whose norm is computed.  Not modified.
     */
    public static double normInf( DenseMatrix64F a ) {
        return CommonOps.maxAbs(a);
    }

    /**
     * Sets the values inside the matrix so that it is an identity matrix.
     *
     * @see #identity(int)
     *
     * @param mat A square matrix.
     */
    public static void setIdentity( DenseMatrix64F mat )
    {
        if( mat.numRows != mat.numCols )
            throw new IllegalArgumentException("Must be a square matrix.");

        int width = mat.numCols;

        double data[] = mat.data;
        int length = mat.getNumElements();

        for( int i = 0; i < length; i++ ) {
            data[i] = 0;
        }

        int index = 0;
        for( int i = 0; i < width; i++ ) {
            data[index] = 1;
            index += width + 1;
        }
    }

    /**
     * <p>
     * Creates an identity matrix of the specified size.<br>
     * <br>
     * a<sub>ij</sub> = 0   if i &ne; j<br>
     * a<sub>ij</sub> = 1   if i = j<br>
     * </p>
     *
     * @param width The width and height of the identity matrix.
     * @return A new instance of an identity matrix.
     */
    public static DenseMatrix64F identity( int width )
    {
        DenseMatrix64F ret = new DenseMatrix64F(width,width);

        for( int i = 0; i < width; i++ ) {
            ret.set(i,i,1.0);
        }

        return ret;
    }

    /**
     * <p>
     * Creates a new matrix whose diagonal elements are specified by diagEl and all
     * the other elements are zero.<br>
     * <br>
     * a<sub>ij</sub> = 0         if i &le; j<br>
     * a<sub>ij</sub> = diag[i]   if i = j<br>
     * </p>
     *
     * @param diagEl Contains the values of the diagonal elements of the resulting matrix.
     * @return A new matrix.
     */
    public static DenseMatrix64F diag( double ...diagEl )
    {
        DenseMatrix64F ret = new DenseMatrix64F(diagEl.length,diagEl.length);

        for( int i = 0; i < diagEl.length; i++ ) {
            ret.set(i,i,diagEl[i]);
        }

        return ret;
    }

    /**
     * <p>
     * Computes the sum of all the elements inside the matrix:<br>
     * <br>
     * &sum;<sub>i=0:m-1,j=0:n-1</sub> a<sub>ij</sub>
     * </p>
     * 
     * @param a A matrix.
     * @return The sum of all the elements in the matrix.
     */
    public static double sumMatrix( DenseMatrix64F a ) {
        double data[] = a.data;
        int length = a.getNumElements();

        double ret = 0;

        for( int i = 0; i < length; i++ ) {
            ret += data[i];
        }

        return ret;
    }

    /**
     * <p>
     * Sets 'sub' equal to the specified submatrix of 'orig'.  Sub is
     * reshaped if neccisary.  The submatrix will contains columns x0 to x1 and rows y0 to y1, all
     * inclusive.
     * </p>
     * <p>
     * s<sub>i-y0 , j-x0</sub> = o<sub>ij</sub> for all y0 &le; i &le; y1 and x0 &le; j &le; x1 <br>
     * <br>
     * where 's<sub>ij</sub>' is an element in the submatrix and 'o<sub>ij</sub>' is an element in the
     * original matrix.
     * </p>
     *
     * @param orig The original matrix which is to be copied.  Not modified.
     * @param x0 Start column.
     * @param x1 Stop column.
     * @param y0 Start row.
     * @param y1 Stop row.
     * @param sub Where the submatrix are stored.  Modified.
     */
    public static void submatrix( DenseMatrix64F orig ,
                                  int x0 , int x1 ,
                                  int y0 , int y1 ,
                                  DenseMatrix64F sub )
    {
        sub.reshape(y1-y0+1,x1-x0+1);
        int y_sub = 0;
        for( int y = y0; y <= y1; y++ ) {
            int x_sub = 0;
            for( int x = x0; x <= x1; x++ ,x_sub++) {
                double v = orig.get(y,x);
                sub.set(y_sub,x_sub,v);
            }
            y_sub++;
        }
    }

    /**
     * Computes the Lower Upper (LU) decomposition of the matrix.
     *
     * @param a A matrix. Not modified.
     * @return LUDecomposition
     */
    public static LUDecomposition lu( DenseMatrix64F a ) {
        LUDecomposition alg = DecompositionFactory.lu(a.numRows);
        alg.decompose(a);
        return alg;
    }

    /**
     * Computes the Singular Value Decomposition (SVD) of the matrix.
     *
     * @param a A matrix. Not modified.
     * @return SvdNumericalRecipes
     */
    public static SingularValueDecomposition svd( DenseMatrix64F a ) {
        SingularValueDecomposition alg = DecompositionFactory.svd(a.numRows,a.numCols);
        alg.decompose(a);
        return alg;
    }

    /**
     * Computes the QR decomposition of the matrix.
     *
     * @param a A matrix. Not modified.
     * @return QRDecomposition
     */
    public static QRDecomposition qr( DenseMatrix64F a ) {
        QRDecomposition alg = DecompositionFactory.qr(a.numRows, a.numCols);
        alg.decompose(a);
        return alg;
    }

    /**
     * Computes the Cholesky decomposition of the matrix.
     *
     * @param a A matrix. Not modified.
     * @return CholeskyDecomposition
     */
    public static CholeskyDecomposition cholesky( DenseMatrix64F a ) {
        CholeskyDecomposition alg = DecompositionFactory.chol(a.numRows);
        alg.decompose(a);
        return alg;
    }
}
